"""
Contains the code to fetch the informations from suse.com
"""
import re
from typing import Dict, DefaultDict
from pathlib import Path
from datetime import timedelta
from collections import defaultdict
from requests_cache import CachedSession
from bs4 import BeautifulSoup
from scf.config import settings
from scf.models import CVE, SimplifiedRating, CVSS, CVSSVector, State, Product, OverallState
from scf.utils import numeric


SUSE_CVE_LIST_URL = 'https://www.suse.com/security/cve/index.html'
SUSE_CVE_URL_TEMPLATE = 'https://www.suse.com/security/cve/{cve}.html'


class Cache:
    """
    Holds the session instance
    """
    _INSTANCE = None

    @staticmethod
    def instance():
        if Cache._INSTANCE is None:
            Cache._INSTANCE = CachedSession(
                str(Path(settings.cache.path) / 'scf.sqlite'),
                expire_after=timedelta(days=settings.cache.expiration_days),
                urls_expire_after={f"*/{cve}.html": -1 for cve in settings.cache.expiration_exceptions},
                backend='sqlite',
                stale_if_error=True)
        return Cache._INSTANCE


def url_to_soup(url: str, timeout: int = 30, use_cache: bool = True):
    """
    Fetches the html and returns a beautifulsoup instance
    """

    cache = Cache.instance()

    if not use_cache:
        cache.cache.delete_url(url)

    user_agent = {'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:100.0) Gecko/20100101 Firefox/100.0'}

    res = cache.get(url,
                    headers=user_agent,
                    timeout=timeout,
                    verify=settings.connection.ssl_verify)

    res.raise_for_status()

    return BeautifulSoup(res.text, 'html.parser')


def list_cve_by_year(timeout: int = 30, use_cache: bool = False) -> Dict[str, list]:
    """
    Fetches all cves grouped by year

    Returns a dict with the years as keys and the cves as values in a list
    """
    soup = url_to_soup(SUSE_CVE_LIST_URL, timeout=timeout, use_cache=use_cache)
    current_year = ''
    cve_list: DefaultDict[str, list] = defaultdict(list)
    for tag in soup.select('div#mainbody > h3, div#mainbody > a'):
        if tag.name == "h3":
            current_year = tag.text.split()[0]
        elif tag.name == "a":
            cve_list[current_year].append(tag.text)
        else:
            raise ValueError(f"Unexpected html tag found (tag: {tag.name})")
    return cve_list


def get_all_cve(timeout: int = 30, use_cache: bool = False) -> list:
    """
    Fetches all cves
    """
    by_year = list_cve_by_year(timeout=timeout, use_cache=use_cache)
    return sorted([cve for ylist in by_year.values() for cve in ylist], key=numeric, reverse=True)


def prefetch_cve(cve: str, timeout: int = 30) -> str:
    """
    Prefetches the cve and save the html into cache
    """
    url = SUSE_CVE_URL_TEMPLATE.format(cve=cve)
    url_to_soup(url, timeout=timeout, use_cache=True)
    return cve


def get_cve_details(cve: str, timeout: int = 30, use_cache: bool = True) -> CVE:
    """
    Fetches the details of a cve

    Returns a CVE object
    """
    url = SUSE_CVE_URL_TEMPLATE.format(cve=cve)
    soup = url_to_soup(url, timeout=timeout, use_cache=use_cache)

    data = {'name': cve, 'url': url}

    # parse description
    description_header_tag = soup.select_one('div#mainbody > h4:-soup-contains("Description")')
    data['description'] = description_header_tag.next_sibling.text.strip('\n')

    # parse overall state
    match = re.search(r"Overall state of this security issue: (\w+)", soup.text)
    if match:
        data['overall_state'] = OverallState[match.groups()[0].upper().replace(' ', '_')]

    # parse suse rating
    match = re.search(r"This issue is currently rated as having (\w+) severity", soup.text)
    if match:
        data['simplified_rating'] = SimplifiedRating[match.groups()[0].upper()]

    # parse cvss
    cvss_table_caption_tag = soup.select_one('div#mainbody > table > caption > a:-soup-contains("CVSS v3 Scores")')
    if cvss_table_caption_tag:
        cvss_table = cvss_table_caption_tag.find_parent('table')

        # parse the cvss data
        base_score = float(list(cvss_table.select_one('td:-soup-contains("Base Score")').parent.children)[-1].text)
        vector = list(cvss_table.select_one('td:-soup-contains("Vector")').parent.children)[-1].text
        version = list(cvss_table.select_one('td:-soup-contains("CVSSv3 Version")').parent.children)[-1].text

        data['cvss'] = CVSS(score=base_score, vector=CVSSVector.from_string(vector), version=version)

    # parse products
    products_table_header_tag = soup.select_one('div#mainbody > table th:-soup-contains("Source package")')
    if products_table_header_tag:
        products_table = products_table_header_tag.find_parent('table')
        products = []

        for row in products_table.select('tr')[1:]:
            product, source, status = [i.text for i in row.select('td')]
            for single_product in product.split('\n'):
                products.append(Product(name=single_product, package=source, state=State[status.upper().replace(' ', '_').replace('\'', '_')]))
        data['affected_products'] = products

    return CVE(**data)



"""
Small api server
"""
import uvicorn
from fastapi import FastAPI
from .models import CVE
from .suse import get_all_cve, get_cve_details, list_cve_by_year, Cache


app = FastAPI(docs_url='/')


@app.post('/cache/clear', summary='Clear the cache', tags=['Cache'])
async def cache_clear():
    """
    Clear the cache
    """
    Cache.instance().cache.clear()
    return {'result': 'success'}, 200


@app.post('/cache/clean', summary='Clean the cache (remove expired entries)', tags=['Cache'])
async def cache_clean():
    """
    Clean the cache
    """
    Cache.instance().remove_expired_responses()
    return {'result': 'success'}, 200


@app.get('/cve', summary='Fetch all known CVEs', tags=['CVE'])
async def cve(use_cache: bool = False):
    """
    Returns a list with all cves
    """
    return get_all_cve(use_cache=use_cache)


@app.get('/cve/year', summary='Fetch all known CVEs grouped by year', tags=['CVE'])
async def cve_by_years(use_cache: bool = False):
    """
    Returns a list with all cves grouped by year
    """
    return list_cve_by_year(use_cache=use_cache)


@app.get('/cve/year/{year}', summary='Fetch all known CVEs for a specific year', tags=['CVE'])
async def cve_by_year(year: str, use_cache: bool = False):
    """
    Returns a list with all cves in the given year
    """
    year_cves = list_cve_by_year(use_cache=use_cache)
    if year in year_cves:
        return year_cves[year]
    return []


@app.get('/cve/{cve_id}', summary='Fetch details for a specific CVE', tags=['CVE'], response_model=CVE)
async def cve_by_id(cve_id: str, use_cache: bool = True):
    """
    Returns the details for a given cve
    """
    return get_cve_details(cve_id, use_cache=use_cache)


def run(host: str = '0.0.0.0',
        port: int = 8000,
        workers: int = 1) -> None:
    """
    Starts the uvicorn server
    """
    uvicorn.run("scf.server:app", host=host, port=port, debug=False, workers=workers)



"""
Contains config specific stuff
"""

from pathlib import Path
from dynaconf import Dynaconf

settings = Dynaconf(
    envvar_prefix="SCF",
    settings_files=['settings.toml',
                    f'{Path.home() / ".config" / "scf" / "settings.toml"}',
                    f'{Path(__file__).parent / "settings.toml"}'],
)



"""
Contains helper functions
"""
import os
import operator
from typing import Tuple
from pathlib import Path
from functools import reduce


def find_by_path(element: str, json: dict):
    """
    Finds a value by a gived key-path
    src: https://stackoverflow.com/a/31033676
    """
    return reduce(operator.getitem, element.split('.'), json)


def file_size(path: Path, human_readable: bool = True) -> str:
    """
    Gets the filesize and optionally formats it
    """
    size_in_bytes = os.path.getsize(path)

    if not human_readable:
        return str(size_in_bytes)

    # https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(size_in_bytes) < 1024.0:
            return f'{size_in_bytes:3.1f}{unit}'
        size_in_bytes /= 1024.0

    return f'{size_in_bytes:3.1f}Yi'


def numeric(cve: str) -> Tuple[int, int]:
    """
    Returns a numeric cve number for sorting
    """
    year, num = cve[4:].split('-')
    return (int(year), int(num))



"""
Contain models used by this package
"""
import json
from typing import List, Optional
from enum import Enum
from pydantic import BaseModel
from rich.tree import Tree
from fastapi.encoders import jsonable_encoder


class OverallState(str, Enum):
    """
    Overall state by Suse: https://www.suse.com/c/cve-pages-self-help-security-issues-suse-linux-enterprise/
    """
    RESOLVED = 'RESOLVED'
    DOES = 'DOES_NOT_AFFECT_SUSE'
    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    ANALYSIS = 'ANALYSIS'
    NEW = 'NEW'
    POSTBONED = 'POSTBONED'
    IGNORE = 'IGNORE'
    REVISIT = 'REVISIT'

    def pretty(self):
        """
        Returns the colorized enum value
        """
        color = {
            'RESOLVED': 'green',
            'DOES_NOT_AFFECT_SUSE': 'green',
            'PENDING': 'yellow',
            'RUNNING': 'slate_blue1',
            'ANALYSIS': 'slate_blue1',
            'REVISIT': 'slate_blue1',
            'NEW': 'slate_blue1',
            'POSTBONED': 'slate_blue1',
            'IGNORE': 'green',
        }[self.name]
        return f'[{color}]{self.value}[/{color}]'


class SimplifiedRating(str, Enum):
    """
    Simplified Rating by Suse: https://www.suse.com/support/security/rating/
    """
    LOW = 'LOW'
    MODERATE = 'MODERATE'
    IMPORTANT = 'IMPORTANT'
    CRITICAL = 'CRITICAL'

    def pretty(self):
        """
        Returns the colorized enum value
        """
        color = {
            'LOW': 'green',
            'MODERATE': 'yellow',
            'IMPORTANT': 'orange_red1',
            'CRITICAL': 'red1',
        }[self.name]
        return f'[{color}]{self.value}[/{color}]'


class State(str, Enum):
    """
    Represents how the state of an affected package can be described
    """
    IN_PROGRESS = 'IN_PROGRESS'
    WON_T_FIX = 'WILL_NOT_FIX'
    UNSUPPORTED = 'UNSUPPORTED'
    RELEASED = 'RELEASED'
    AFFECTED = 'AFFECTED'
    NOT_AFFECTED = 'NOT_AFFECTED'
    ALREADY_FIXED = 'ALREADY_FIXED'
    ANALYSIS = 'ANALYSIS'
    ASK_MAINTAINER = 'ASK_MAINTAINER'
    IGNORE = 'IGNORE'

    def pretty(self):
        """
        Returns the colorized enum value
        """
        color = {
            'NOT_AFFECTED': 'green',
            'ALREADY_FIXED': 'green',
            'RELEASED': 'green',
            'IGNORE': 'green',
            'UNSUPPORTED': 'orange_red1',
            'IN_PROGRESS': 'slate_blue1',
            'AFFECTED': 'red1',
            'ANALYSIS': 'slate_blue1',
            'WON_T_FIX': 'yellow',
            'ASK_MAINTAINER': 'yellow',
        }[self.name]
        return f'[{color}]{self.value}[/{color}]'


class Product(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Contains information about a package that may or maybe not is affected by a given cve
    """
    name: str
    package: str
    state: State

    def tree(self):
        """
        Creates a pretty rich tree object
        """
        t = Tree(self.name)
        p = t.add('Package')
        p.add(self.package)
        s = t.add('State')
        s.add(self.state.pretty())
        return t


class CVSSVector(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Contains the parsed vector information
    """
    _STR2VAR = {
        'av': {
            'name': 'access_vector',
            'values': {
                'N': 'Network',
                'A': 'Adjacent',
                'L': 'Local',
                'P': 'Physical'
            }
        },
        'ac': {
            'name': 'access_complexity',
            'values': {
                'L': 'Low',
                'H': 'High',
            },
        },
        'pr': {
            'name': 'privileges_required',
            'values': {
                'N': 'None',
                'L': 'Low',
                'H': 'High',
            },
        },
        'ui': {
            'name': 'user_interaction',
            'values': {
                'N': 'None',
                'R': 'Required',
            },
        },
        's': {
            'name': 'scope',
            'values': {
                'U': 'Unchanged',
                'C': 'Changed',
            },
        },
        'c': {
            'name': 'confidentiality_impact',
            'values': {
                'H': 'High',
                'L': 'Low',
                'N': 'None',
            },
        },
        'i': {
            'name': 'integrity_impact',
            'values': {
                'H': 'High',
                'L': 'Low',
                'N': 'None',
            },
        },
        'a': {
            'name': 'availability_impact',
            'values': {
                'H': 'High',
                'L': 'Low',
                'N': 'None',
            },
        },
        'cvss': {'name': 'version', 'values': {}},
    }
    raw: str
    access_vector: str
    access_complexity: str
    privileges_required: str
    user_interaction: str
    scope: str
    confidentiality_impact: str
    integrity_impact: str
    availability_impact: str
    version: str

    def tree(self):
        """
        Creates a pretty rich tree object
        """
        t = Tree('[light_slate_blue]Vector')

        t_r = t.add('[sky_blue2]Raw')
        t_r.add(self.raw)

        t_av = t.add('[sky_blue2]Access Vector')
        t_av.add(self.access_vector)

        t_ac = t.add('[sky_blue2]Access Complexity')
        t_ac.add(self.access_complexity)

        t_pr = t.add('[sky_blue2]Privileges Required')
        t_pr.add(self.privileges_required)

        t_ui = t.add('[sky_blue2]User Interaction')
        t_ui.add(self.user_interaction)

        t_s = t.add('[sky_blue2]Scope')
        t_s.add(self.scope)

        t_ci = t.add('[sky_blue2]Confidentiality Impact')
        t_ci.add(self.confidentiality_impact)

        t_ii = t.add('[sky_blue2]Integrity Impact')
        t_ii.add(self.integrity_impact)

        t_ai = t.add('[sky_blue2]Availability Impact')
        t_ai.add(self.availability_impact)

        t_v = t.add('[sky_blue2]Version')
        t_v.add(self.version)

        return t

    @staticmethod
    def from_string(cvss_str):
        """
        Parses a raw cvss vector string
        """
        data = {'raw': cvss_str}
        for name, value in [kv.split(':') for kv in cvss_str.split('/')]:
            lookup = CVSSVector._STR2VAR[name.lower()]
            data[lookup['name']] = lookup['values'].get(value, value)
        return CVSSVector(**data)


class CVSS(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Contains the fetched cvss information
    """
    score: float
    vector: CVSSVector
    version: str

    def tree(self):
        """
        Creates a pretty rich tree object
        """
        t = Tree('[slate_blue1]CVSS')

        t_s = t.add('[light_slate_blue]Score')
        t_s.add(str(self.score))

        t.add(self.vector.tree())

        t_v = t.add('[light_slate_blue]Version')
        t_v.add(self.version)

        return t


class CVE(BaseModel):  # pylint: disable=too-few-public-methods
    """
    Holds all the information about the cve
    """
    name: str
    description: str
    url: str
    overall_state: Optional[OverallState]
    cvss: Optional[CVSS]
    simplified_rating: Optional[SimplifiedRating]
    affected_products: Optional[List[Product]]

    def asdict(self):
        """
        Returns the data as dict
        """
        return dict(jsonable_encoder(self))

    def json(self):
        """
        Returns the data as json
        """
        return json.dumps(self.asdict())

    def tree(self):
        """
        Creates a pretty rich tree object
        """
        t = Tree(f'[sea_green3]{self.name}')
        t_d = t.add('[slate_blue1]Description')
        t_d.add(self.description)

        t_u = t.add('[slate_blue1]Url')
        t_u.add(self.url)

        if self.overall_state:
            t_os = t.add('[slate_blue1]Overall state')
            t_os.add(self.overall_state.pretty())

        if self.cvss:
            t.add(self.cvss.tree())

        if self.simplified_rating:
            t_s = t.add('[slate_blue1]Simplified Rating')
            t_s.add(self.simplified_rating.pretty())

        if self.affected_products:
            t_p = t.add('[slate_blue1]Products')
            for p in self.affected_products:
                t_p.add(p.tree())

        return t



import os
import json
import re
from random import random
from time import sleep
from multiprocessing import Pool
from typing import Optional, cast, Dict, List, Union
from datetime import datetime
from pathlib import Path
from subprocess import call, Popen, DEVNULL
from collections import defaultdict

import typer
import toml
from scf import __version__
from scf.models import OverallState
from scf.server import run as server_start
from scf.suse import get_cve_details, list_cve_by_year, get_all_cve, prefetch_cve
from rich import print
from rich.progress import Progress
from rich.tree import Tree
from rich.live import Live
from rich.table import Table
from scf.suse import Cache
from scf.utils import find_by_path, file_size, numeric
from requests_cache.backends.sqlite import SQLiteCache
from scf.config import settings


app = typer.Typer(context_settings={'help_option_names': ['-h', '--help']})
cache_cmd = typer.Typer(context_settings={'help_option_names': ['-h', '--help']}, add_completion=False)
cve_cmd = typer.Typer(context_settings={'help_option_names': ['-h', '--help']}, add_completion=False)
server_cmd = typer.Typer(context_settings={'help_option_names': ['-h', '--help']}, add_completion=False)
config_cmd = typer.Typer(context_settings={'help_option_names': ['-h', '--help']}, add_completion=False)

app.add_typer(cache_cmd, name='cache')
app.add_typer(cve_cmd, name='cve')
app.add_typer(server_cmd, name='server')
app.add_typer(config_cmd, name='config')


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"v{__version__}")
        raise typer.Exit()


@config_cmd.command("init")
def config_cmd_init(
        overwrite: bool = typer.Option(False, '--overwrite', help='Overwrite if already exists.'),) -> None:
    """
    Creates the default config in the home directory
    """
    target = Path.home() / '.config' / 'scf' / 'settings.toml'
    if target.exists() and not overwrite:
        typer.echo(f'Config already exists ({target})', err=True)
        raise typer.Exit(1)
    os.makedirs(target.parent, exist_ok=True)

    with open(Path(__file__).parent / 'settings.toml', 'rt', encoding='utf-8') as default_config,\
            open(target, 'wt', encoding='utf-8') as target_config:
        target_config.write(default_config.read())


@config_cmd.command("dump")
def config_cmd_dump():
    """
    Dumps the current config
    """
    typer.echo(toml.dumps(settings.as_dict()))


@config_cmd.command("edit")
def config_cmd_edit():
    """
    Edit the current config
    """
    config = Path.home() / '.config' / 'scf' / 'settings.toml'
    if not config.exists():
        config_cmd_init()
    editor = os.environ.get('EDITOR', 'vim')
    raise typer.Exit(call([editor, config]))


@server_cmd.command("run")
def server_cmd_run(
        host: str = typer.Option('0.0.0.0', '--host', help='The host ip to bind the socket to.'),
        port: int = typer.Option(8000, '--port', help='The port to which the socket should listen to.'),
        workers: int = typer.Option(os.cpu_count(), '-w', '--workers', help='The number of workers that should be started'),) -> None:
    """
    Starts the uvicorn server
    """
    server_start(host=host, port=port, workers=workers)


@cache_cmd.command("clean")
def cache_cmd_clean() -> None:
    """
    Remove expired cache entries
    """
    Cache.instance().remove_expired_responses()


@cache_cmd.command("clear")
def cache_cmd_clear() -> None:
    """
    Clear the whole cache
    """
    Cache.instance().cache.clear()


@cache_cmd.command("stats")
def cache_cmd_stats() -> None:
    """
    Get some stats about the cache
    """
    with Progress(transient=True) as progress:
        progress.add_task('Collecting data...', total=None)
        cache = cast(SQLiteCache, Cache.instance().cache)
        path = Path(cache.db_path)
        size = file_size(path)
        all_entries = cache.response_count()
        not_expired = cache.response_count(check_expiry=True)
        expired = all_entries - not_expired

    typer.echo(f'Location: {path}')
    typer.echo(f'Size: {size}')
    typer.echo(f'Total entries: {all_entries}')
    if all_entries > 0:
        typer.echo(f'Expired entries: {expired} ({100/all_entries*expired:.2f}%)')


@cache_cmd.command("populate")
def cache_cmd_populate(
        cve_filter: str = typer.Option(f'CVE-{datetime.now().year}-.*',
                                       '--filter', help="Regex to apply on the CVEs to fetch."),
        workers: int = typer.Option(1, '-w', '--workers', help='The number of workers that should be started'),) -> None:
    """
    Prefetch all the cve data
    """

    re_filter = re.compile(cve_filter)
    cve_list = list(filter(re_filter.match, get_all_cve(use_cache=False)))

    with Progress(transient=True) as progress:
        task = progress.add_task('Populating cache...', total=len(cve_list))

        with Pool(processes=workers) as pool:
            for cve in pool.imap_unordered(prefetch_cve, cve_list):
                progress.update(task, description=f'Populating cache...{cve}')
                progress.advance(task)
                sleep(random())


@cve_cmd.command("details")
def cve_cmd_details(
        name: str = typer.Argument(None, help='The cve identifier to lookup.'),
        disable_cache: bool = typer.Option(False, '--no-cache', help='Disable the cache.'),
        field: str = typer.Option(None, '--field', help='Only print the given field (e.g. `cvss.score`)'),
        use_json: bool = typer.Option(False, '--json', help='Print the result as json.'),) -> None:
    """
    Fetch the details of a given cve
    """
    details = get_cve_details(name, use_cache=not disable_cache)
    if field is not None:
        selected = find_by_path(field, details.asdict())
        if use_json:
            typer.echo(json.dumps(selected))
        typer.echo(selected)
    elif use_json:
        typer.echo(details.json())
    else:
        print(details.tree())


@cve_cmd.command("watch")
def cve_cmd_watch(
        command: str = typer.Option(None, '-c', '--command', help='Run this command if a new CVE is found'),
        test: bool = typer.Option(False, '-t', '--test', help='Test the command and exit'),
        interval: int = typer.Option(30, '-i', '--interval', help='Set the refresh interval'),) -> None:
    """
    Show the latest CVE as they get published
    """
    if test and not command:
        typer.echo('Give me a command.', err=True)
        raise typer.Exit(1)
    elif test:
        rc = call([command], shell=True, stdout=DEVNULL)
        typer.echo(f'Executed command. RC = {rc}')
        raise typer.Exit(0)

    table = Table()
    table.add_column('Time')
    table.add_column('CVE')
    table.add_column('Rating')
    table.add_column('Score')
    table.add_column('Overall state')

    last_cve = None
    first = True

    with Live(table, auto_refresh=False) as live:
        while True:
            try:
                all_cve = get_all_cve()
                new_cve = []

                if last_cve is None:
                    last_cve = all_cve[0]
                    new_cve.append(last_cve)
                else:
                    while all_cve[0] != last_cve:
                        new_cve.append(all_cve[0])
                        all_cve.pop()

                for cve in reversed(new_cve):
                    cve_data = get_cve_details(cve)
                    rating = cve_data.simplified_rating or 'Unknown'
                    overall = cve_data.overall_state or 'Unknown'

                    if cve_data.cvss is not None and cve_data.cvss.score is not None:
                        score = str(cve_data.cvss.score)
                    else:
                        score = 'Unknown'

                    table.add_row(datetime.now().strftime('%Y/%m/%d %H:%M:%S'), cve, rating, score, overall)
                    live.update(table, refresh=True)
                    if command and not first:
                        Popen([command], shell=True, stdout=DEVNULL)
                sleep(interval)
                first = False
            except KeyboardInterrupt:
                break


@cve_cmd.command("list")
def cve_cmd_list(
        grouped: bool = typer.Option(False, '--grouped-by-year', help='List cve grouped by year.'),
        year: int = typer.Option(None, '-y', '--year', help='List cve for a specific year.'),
        unresolved: bool = typer.Option(False, '-u', '--unresolved-only', help='Only print cve which are not resolved yet.'),
        use_cache: bool = typer.Option(False, '--cache', help='Enables the cache.'),
        use_json: bool = typer.Option(False, '--json', help='Print the result as json.'),) -> None:
    """
    Fetch the known cve for a given year or all.
    """
    cves: Union[Dict, List] = []

    if grouped:
        cves = list_cve_by_year(use_cache=use_cache)
    elif year is not None:
        cves = list_cve_by_year(use_cache=use_cache)[str(year)]
    else:
        cves = get_all_cve(use_cache=use_cache)

    # apply the filters
    if unresolved:
        if isinstance(cves, dict):
            filtered = defaultdict()
            for cve_year, cve_list in cves.items():
                for cve in cve_list:
                    details = get_cve_details(cve, use_cache=True)
                    if details.overall_state != OverallState.RESOLVED:
                        filtered[cve_year].append(cve)
        else:
            filtered = []
            for cve in cves:
                details = get_cve_details(cve, use_cache=True)
                if details.overall_state != OverallState.RESOLVED:
                    filtered.append(cve)

        cves = filtered

    if use_json:
        typer.echo(json.dumps(cves))
    else:
        tree = Tree('CVE')
        if isinstance(cves, dict):
            for cve_year, cve_list in cves.items():
                t_y = tree.add(cve_year)
                for cve in sorted(cve_list, key=numeric, reverse=True):
                    t_y.add(cve)
        else:
            for cve in cves:
                tree.add(cve)
        print(tree)


@app.callback()
def main(
    _: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        help="Show the application's version and exit.",
        callback=_version_callback,
        is_eager=True,)) -> None:
    """
    scf fetches informations about CVEs from suse.com.
    """
    return
